#include "Hit.h"
#include "TMath.h"
#include "TRandom.h"
#include <cmath>

#ifdef USER_TUPLE
#include "TNtupleD.h"
#include <sstream>
#endif

ClassImp(TPC::Hit)
ClassImp(TPC::Hit::Pulse)
ClassImp(TPC::Hit::Pulse::VGetTimeMethod)
ClassImp(TPC::Hit::Pulse::Cog)
ClassImp(TPC::Hit::Pulse::Inflection)


namespace TPC {
bool Hit::Pulse::fgWithNoise = kTRUE;
bool Hit::Pulse::fgWithADCLimit = kTRUE;
#ifdef USER_TUPLE
TNtupleD *Hit::fgTup = 0;
#endif
// C-tor
Hit::Hit(Track &trk,
         GEM   &gem,
         Pasa  &pca)
   : fTrackPtr(&trk),
     fGEMPtr  (&gem),
     fPasaPtr (&pca)
{
    //// FIXME ////////////////////////////////////////
#if 0
    static const int kNp = 5; // pad range (-kNp,+kNp)
#else
#if 0 //2021/03/23
    static const int kNp = 7; // pad range (-kNp,+kNp)
#else
    static const int kNp = 10; // pad range (-kNp,+kNp)
#endif
#endif
    ///////////////////////////////////////////////////
    for (int a=-kNp; a<=kNp; a++) {
        Add(new Pulse(a,&pca));
    }
    SetOwner(); // Hit owns "Pulse"s

    // FIXME
    ///////////////////////////////////////////////////////////
    static const double kMinQ   = 1.e-6; // minimum charge/pad
    static const double kNe2ADC = pca.GetNorm(); // #e's to adc
    ///////////////////////////////////////////////////////////
    double vd   = trk.GetVdrift()/1000.; // [mm/nsec]
#if 0
    double d0   = trk.GetXv0().Z();
    double t0   = d0/vd;
    int    it0  = t0/pca.GetW();
    double toff = t0 - it0*pca.GetW();
#else
    double toff = 0.;
#endif
#ifdef USER_TUPLE
    int    ksum   = 0;
    int    k2sum  = 0;
    double gsum   = 0.;
    double g2sum  = 0.;
    double qsum   = 0.;
    double q2sum  = 0.;
    double qc2sum = 0.;
#endif
    TIter next(&trk);
    Track::Cluster *clp;
    //--
    // loop oever primary clusters.
    //--
    while ((clp = static_cast<Track::Cluster *>(next()))) {
        TVector3 xv0  = clp->GetXv0();
        TIter nexte(clp);
	TVector3 *xvp;
        //--
        // loop over electrons in this cluster.
        //--
        int    kc = 0;
        double qc = 0.;
        while ((xvp = static_cast<TVector3 *>(nexte()))) {
            TVector3 &xv = *xvp;
            double x = xv.X();
            double y = xv.Y();
            double z = xv.Z();                // relative to track
            double t = z/vd + toff;           // t is measured from track
            double g = kNe2ADC*gem.GasGain(); // gas gain for this electron [adc counts]
            //--
	    // loop over pads to add a signal by this electron.
            //--
	    for (int a=-kNp; a<=kNp; a++) {
		double q     = g*gem.Fx(a,x,y); // charge on this pad
		if (q < kMinQ) continue;        // threshold
#ifdef USER_TUPLE
                if (!a) {
                    kc++;
                    qc    += q;
                    ksum++;
                    gsum  += g;
                    g2sum += g*g;
                    qsum  += q;
                    q2sum += q*q;
                }
#endif
                int    indx  = a + kNp;
		Pulse &puls  = *static_cast<Pulse *>((*this)[indx]);
		Pulse  pul(t,q,a,pca); // pulse by this electron on pad "a"
		       puls += pul;    // add this pulse
	    }
	}
        k2sum  += kc*kc;   
        qc2sum += qc*qc;
    }
    //--
    // Loop over pads and add noise and apply threshold.
    //--
    for (int a=-kNp; a<=kNp; a++) {
        int    indx  = a + kNp;
        Pulse &puls  = *static_cast<Pulse *>((*this)[indx]);
        if (puls.GetWithNoise()) puls.AddNoise();
        if (puls.GetWithADCLimit()) puls.LimitADC();
    }
#ifdef USER_TUPLE
    static const int kMin = Pulse::fgMin;
    double nc       = trk.GetEntries();
    double gav      = gsum/ksum;
    double g2av     = g2sum/ksum;
    double qav      = qsum/ksum;
    double q2av     = q2sum/ksum;
    double neffinv  = g2sum/(gsum*gsum);
    double k2sumav  = k2sum/nc;
    double qc2sumav = qc2sum/nc;
    double nheffinv = qc2sum/(qsum*qsum);
    double qm2     = static_cast<Pulse *>((*this)[-2+kNp])->GetCharge();
    double qm1     = static_cast<Pulse *>((*this)[-1+kNp])->GetCharge();
    double q0      = static_cast<Pulse *>((*this)[ 0+kNp])->GetCharge();
    double qp1     = static_cast<Pulse *>((*this)[+1+kNp])->GetCharge();
    double qp2     = static_cast<Pulse *>((*this)[+2+kNp])->GetCharge();
    double data[36];
    data[ 0] = gav;
    data[ 1] = g2av;
    data[ 2] = qav;
    data[ 3] = q2av;
    data[ 4] = neffinv;
    data[ 5] = ksum;
    data[ 6] = nc;
    data[ 7] = k2sumav;
    data[ 8] = qc2sumav;
    data[ 9] = nheffinv;
    data[10] = qm2;
    data[11] = qm1;
    data[12] = q0;
    data[13] = qp1;
    data[14] = qp2;
    data[15] = (*static_cast<Pulse *>((*this)[kNp]))[-1-kMin];
    data[16] = (*static_cast<Pulse *>((*this)[kNp]))[+0-kMin];
    data[17] = (*static_cast<Pulse *>((*this)[kNp]))[+1-kMin];
    data[18] = (*static_cast<Pulse *>((*this)[kNp]))[+2-kMin];
    data[19] = (*static_cast<Pulse *>((*this)[kNp]))[+3-kMin];
    data[20] = (*static_cast<Pulse *>((*this)[kNp]))[+4-kMin];
    data[21] = (*static_cast<Pulse *>((*this)[kNp]))[+5-kMin];
    data[22] = (*static_cast<Pulse *>((*this)[kNp]))[+6-kMin];
    data[23] = (*static_cast<Pulse *>((*this)[kNp]))[+7-kMin];
    data[24] = (*static_cast<Pulse *>((*this)[kNp]))[+8-kMin];
    data[25] = (*static_cast<Pulse *>((*this)[kNp]))[+9-kMin];
    data[26] = (*static_cast<Pulse *>((*this)[kNp]))[+10-kMin];
    data[27] = (*static_cast<Pulse *>((*this)[kNp]))[+11-kMin];
    data[28] = (*static_cast<Pulse *>((*this)[kNp]))[+12-kMin];
    data[29] = (*static_cast<Pulse *>((*this)[kNp]))[+13-kMin];
    data[30] = (*static_cast<Pulse *>((*this)[kNp]))[+14-kMin];
    data[31] = (*static_cast<Pulse *>((*this)[kNp]))[+15-kMin];
    data[32] = (*static_cast<Pulse *>((*this)[kNp]))[+16-kMin];
    data[33] = (*static_cast<Pulse *>((*this)[kNp]))[+17-kMin];
    data[34] = (*static_cast<Pulse *>((*this)[kNp]))[+18-kMin];
    data[35] = (*static_cast<Pulse *>((*this)[kNp]))[+19-kMin];
    GetTupInstancePtr()->Fill(data);
#endif
}

#ifdef USER_TUPLE
TNtupleD *Hit::GetTupInstancePtr()
{
    if (!fgTup) {
        std::stringstream sout;
        sout << "gav:g2av:qav:q2av:neffinv:ksum:nc:k2sumav:qc2sumav:nheffinv:qm2:qm1:qm0:qp1:qp2:"
             << "am1:a0:a1:a2:a3:a4:a5:a6:a7:a8:a9:a10:a11:a12:a13:a14:a15:a16:a17:a18:a19" 
	     << std::ends;
        fgTup = new TNtupleD("tup","",sout.str().data());
    }
    return fgTup;
}

void Hit::ResetTup()
{
    delete fgTup;
    fgTup = 0;
}
#endif

double Hit::GetX(int n) // from c.o.g.
{
    // Charge centroid.
    double w    = fGEMPtr->GetPadWidth();
    double xcog = 0.;
    double qsum = 0.;
    TIter next(this);
    Pulse *pp;
    while ((pp = static_cast<Pulse *>(next()))) {
        Pulse &p = *pp;
	double a = p.GetChannel();
        if (TMath::Abs(a) > n) continue; // avoid side pads
        double q = p.GetCharge();
        xcog += q*a;
	qsum += q;
    }
    xcog *= w/qsum;
    xcog += fGEMPtr->GetOffset();
    return xcog;
}

double Hit::GetZ(int n) // from charge-weighted time average
{
    double vd   = fTrackPtr->GetVdrift()/1000; // [mm/ns]
    double zcog = 0.;
    double qsum = 0.;
    TIter next(this);
    Pulse *pp;
    while ((pp = static_cast<Pulse *>(next()))) {
        Pulse &p = *pp;
	double a = p.GetChannel();
        if (TMath::Abs(a) > n) continue; // avoid side time slots
        double q = p.GetCharge();
	if (q <= 0.) continue;
        zcog += q*p.GetTime();
	qsum += q;
    }
    zcog /= qsum;
    zcog *= vd;
    return zcog;
}

double Hit::GetQ(int n)
{
    double qsum = 0.;
    TIter next(this);
    Pulse *pp;
    while ((pp = static_cast<Pulse *>(next()))) {
        Pulse &p = *pp;
	double a = p.GetChannel();
        if (TMath::Abs(a) > n) continue; // avoid side pads
        double q = p.GetCharge();
	qsum += q;
    }
    return qsum;
}


//-----------------------------------------------------------------------------
// Sub-classes
//-----------------------------------------------------------------------------
const int Hit::Pulse::fgMin = -10; // minimum time bin #
const int Hit::Pulse::fgMax = +40; // maximum time bin #

Hit::Pulse::VGetTimeMethod * Hit::Pulse::fgGetTimeMethodPtr = 0;

// C-tors
Hit::Pulse::Pulse(int    a,
                  Pasa  *pp)
   : TVectorD(fgMax-fgMin+1),
     fCh     (a),
     fCharge (0.),
     fPasaPtr(pp)
{
}

Hit::Pulse::Pulse(double t,
                  double q,
                  int    a,
                  Pasa  &pca)
   : TVectorD(fgMax-fgMin+1),  // ADC data
     fCh     (a),              // channel ID = pad#
     fCharge (q),              // charge on this pad
     fPasaPtr(&pca)            // pointer to the PASA object
{
    const int kMax = fgMax - fgMin;
    for (int i=0; i<kMax; i++) {      // loop over time bins
        int b = i + fgMin;            // time bin #
	double adc = q*pca.Ft(t,b);   // ADC value for this time bin
        (*this)[i] = adc;
    }
}

double Hit::Pulse::GetTime()
{
    return GetGetTimeMethodPtr()->GetTime(*this);
}

Hit::Pulse::VGetTimeMethod *Hit::Pulse::GetGetTimeMethodPtr()
{
    if (!fgGetTimeMethodPtr) {
        fgGetTimeMethodPtr = new Inflection();
    }
    return fgGetTimeMethodPtr;
}

void Hit::Pulse::SetGetTimeMethodPtr(VGetTimeMethod *p)
{
    if (p && fgGetTimeMethodPtr) delete fgGetTimeMethodPtr;
    fgGetTimeMethodPtr = p;
}

Hit::Pulse Hit::Pulse::operator+(Hit::Pulse &p2)
{
    if (p2.fCh != fCh) return *this; // pulses on different pads do not add up
    const int kMax = fgMax - fgMin;
    Pulse sum(fCh,fPasaPtr);
    for (int i=0; i<kMax; i++) {
        double adc1 = (*this)[i];
        double adc2 = p2[i];
        sum[i] = adc1 + adc2;
    }
    sum.fCharge = fCharge + p2.fCharge;
    return sum;
}

Hit::Pulse &Hit::Pulse::operator+=(Hit::Pulse &p2)
{
    const int kMax = fgMax - fgMin; // pulses on different pads do not add up
    for (int i=0; i<kMax; i++) {
        if (p2.fCh != fCh) continue;
        (*this)[i] += p2[i];
    }
    fCharge += p2.fCharge;
    return *this;
}

void Hit::Pulse::AddNoise()
{
    //FIXME//////////////////////////////////////////////
    static const double kNoise  = fPasaPtr->GetNoise();
    static const double kThresh = 5.*kNoise; // 5-sigma
    // should add common mode noise, too.
    /////////////////////////////////////////////////////
    const int kMax = fgMax - fgMin;
    double qsum = 0.;
#if 0
    for (int i=0; i<kMax; i++) {      // loop over time bins
        (*this)[i] += gRandom->Gaus(kNoise);
#else // baseline fluctuation
    double noise = gRandom->Gaus(kNoise);
    for (int i=0; i<kMax; i++) {      // loop over time bins
        (*this)[i] += noise;
#endif
        if ((*this)[i] < kThresh) {
            (*this)[i] = 0.;
            continue;
        }
        qsum += (*this)[i];
	//std::cerr << "th=" << kThresh << " a=" << fCh << " b=" << i + fgMin << " adc=" << (*this)[i] << std::endl;
    }
    fCharge = qsum; // update charge.
}

void Hit::Pulse::LimitADC()
{
    //FIXME//////////////////////////////////////////////
    static const double kADCMax  = 923;
    /////////////////////////////////////////////////////
    const int kMax = fgMax - fgMin;
    double qsum = 0.;

    for (int i=0; i<kMax; i++) {      // loop over time bins
        if ((*this)[i] > kADCMax) (*this)[i] = kADCMax;
        qsum += (*this)[i];
    }
    fCharge = qsum; // update charge.
}

Hit::Pulse::Cog::Cog(Pasa *pcap)
      : VGetTimeMethod(pcap)
{
    if (!pcap) {
        fPasaPtr = new Pasa();
    } 
}

// ----------------
// Utility function
// ----------------
int FindPeak(int n, double data[])
{
    double pkval = -1.e20;
    int    ipk   = -99999;
    for (int i=0; i<n; i++) {
       if (data[i] > pkval) {
	       pkval = data[i];
	       ipk   = i;
       } 
    }
    return ipk;
}

double Hit::Pulse::Cog::GetTime(TVectorD &data)
{
    //--
    // C.o.g using all time bins.
    //--
    double tcog = 0.;
    double qsum = 0.;
    const int kMin = Pulse::GetTimeBinMin();
    const int kMax = Pulse::GetTimeBinMax() - kMin;
#if 0
    // FIXME ////////////////////////////////////////////////////
    // Implement peak hunting and restriction of time bin range
    for (int i=0; i<kMax; i++) {
    /////////////////////////////////////////////////////////////
#else
    /////////////////////
    const int kNhalf = 8; //2->8 20210520
    /////////////////////
    int ipk  = FindPeak(kMax,data.GetMatrixArray());
    int imin = TMath::Max(0,ipk-kNhalf);
    int imax = TMath::Min(kMax,ipk+kNhalf+1);
    for (int i=imin; i<imax; i++) {
#endif
        int    b   = i + kMin;
	double adc = data[i];
        tcog += adc*b;
	qsum += adc;
    }
    tcog *= fPasaPtr->GetW()/qsum;
    tcog -= fPasaPtr->GetTpeak();  // correct for the peaking time delay
    return tcog;
}

Hit::Pulse::Inflection::Inflection(Pasa *pcap)
      : VGetTimeMethod(pcap)
{
    if (!pcap) {
        fPasaPtr = new Pasa();
    } 
}

double Hit::Pulse::Inflection::GetTime(TVectorD &data)
{
    //FIXME///////////////////////////////////////////////////////////////
    // 4-sigma cut
    //static const double kThresh = 4.*TMath::Sqrt(2.)*fPasaPtr->GetNoise();
    static const double kThresh = 2.*TMath::Sqrt(2.)*fPasaPtr->GetNoise();
    //static const double kThresh = 1.*fPasaPtr->GetNoise();
    //////////////////////////////////////////////////////////////////////
    //--
    // Inflection point.
    //--
    const int kMin = Pulse::GetTimeBinMin();
    const int kMax = Pulse::GetTimeBinMax() - kMin;
    double del[kMax];
    double delmax = -1.e20;
    int    imax   = 0;
    for (int i=0; i<kMax-1; i++) {
        if (data[i] < kThresh && data[i+1] < kThresh) {
           del[i] = 0.;
	   continue;
        }
        double diff = data[i+1] - data[i];
	del[i] = diff;
	//std::cerr << "i=" << i << " diff=" << diff << std::endl;
	if (diff > delmax) {
            delmax = diff;    
	    imax   = i;
	}
    }
    double tcog = 0.;
    double dsum = 0.;
#if 0
    // FIXME ////////////////////////////////////////////////////
    // Implement peak hunting and restriction of time bin range
    /////////////////////////////////////////////////////////////
    for (int i=imax-10; i<=imax+2; i++) {
#else
    /////////////////////
    const int kNhalf = 10;
    /////////////////////
    for (int i=imax-kNhalf; i<=imax+kNhalf; i++) {
#endif
	if (i > kMax - 1) break;
	if (i < 0) continue;
	double diff = del[i];
	if (diff < -kThresh) break;
        int    b    = i + kMin;
	//std::cerr << "b=" << b << " diff=" << diff << std::endl;
        tcog += diff*b;
	dsum += diff;
    }
    if (dsum == 0) tcog = (imax + kMin - 0.5)*fPasaPtr->GetW();
    else           tcog *= fPasaPtr->GetW()/dsum;
    //std::cerr << "bmax=" << imax + fgMin << " dsum=" << dsum << " tcog=" << tcog << std::endl;
    tcog -= 0.5*fPasaPtr->GetTpeak();  // correct for the peaking time delay
    return tcog;
}
}
