#ifndef TPC_HIT_H
#define TPC_HIT_H
#include "Track.h"
#include "GEM.h"
#include "Pasa.h"
#include "TObjArray.h"
#include "TVectorD.h"
#ifdef USER_TUPLE
class TNtupleD;
#endif
//-----------------------------------------------------------------------------
// =========
// class Hit 
// =========
// (Description)
//    Class to describe a MPGD Hit.
// (Provides)
//    TPC::Hit               : Hit object
// (Update Record)
//    2015/07/11  K.Fujii     Original version.
//-----------------------------------------------------------------------------
namespace TPC {
class Hit : public TObjArray {
public:
    Hit(Track &trk, GEM &gem, Pasa &pca);

    double GetX(int n = 9999); // use pads in (-n,+n)
    double GetZ(int n = 9999); // use pads in (-n,+n)
    double GetQ(int n = 9999); // use pads in (-n,+n)

#ifdef USER_TUPLE
    static TNtupleD *GetTupInstancePtr();
    static void      ResetTup         ();
#endif

    class Pulse : public TVectorD {
    friend Hit;
    public:
        Pulse(int     a    = -9999,  // Channel ID
              Pasa   *pcap = 0);     // Pasa

        Pulse(double  t,    // arrival time
              double  q,    // charge
              int     a,    // channel ID
	      Pasa   &pca); // Pasa

        inline double GetChannel() { return fCh; }
        inline double GetCharge () { return fCharge; }
	       double GetTime   ();

	inline static int GetTimeBinMin() { return fgMin; }
	inline static int GetTimeBinMax() { return fgMax; }

	inline static bool GetWithNoise   () { return fgWithNoise; }
	inline static bool GetWithADCLimit() { return fgWithADCLimit; }

        class VGetTimeMethod;
        static void             SetGetTimeMethodPtr(VGetTimeMethod *p);
        static VGetTimeMethod  *GetGetTimeMethodPtr();

	inline static void SetWithNoise   (bool set = kTRUE) { fgWithNoise = set; }
	inline static void SetWithADCLimit(bool set = kTRUE) { fgWithADCLimit = set; }

        inline double operator()(int b) { return (*this)[b-fgMin]; }

	Pulse  operator+ (Pulse &p2);
	Pulse &operator+=(Pulse &p2);

	void   AddNoise();
        void   LimitADC(); //Add 20210528

        class VGetTimeMethod {
        public:     
            VGetTimeMethod(Pasa *pcap) : fPasaPtr(pcap) {}
	    virtual double GetTime(TVectorD &data) = 0;

        protected:
            Pasa *fPasaPtr; // readout electcronics

            ClassDef(VGetTimeMethod, 1)   // Inflection GetTime class
        };

        class Cog : public VGetTimeMethod {
        public:
            Cog(Pasa *pcap = 0);

	    double GetTime(TVectorD &data);

        private:
            ClassDef(Cog, 1)   // C.O.G. GetTime class
        };

        class Inflection : public VGetTimeMethod {
        public:
            Inflection(Pasa *pcap = 0);

	    double GetTime(TVectorD &data);

        private:
            ClassDef(Inflection, 1)   // Inflection GetTime class
        };

    private:
        inline void SetChID(int a) { fCh = a; }

    private:        
	             int    fCh;       //  channel ID (pad #)
		     double fCharge;   //  charge
		     Pasa  *fPasaPtr;  //  pointer to PASA
	static const int    fgMin;     //! minimum time bin #
	static const int    fgMax;     //! maximum time bin #
	static       bool   fgWithNoise;     //! w/wo Noise
	static       bool   fgWithADCLimit;     //! w/wo ADC saturation

	static VGetTimeMethod *fgGetTimeMethodPtr; //! pointer to GetTime method

        ClassDef(Pulse, 1)   // Pulse class
    };

private:
    Track *fTrackPtr; // input track
    GEM   *fGEMPtr;   // GEM module
    Pasa  *fPasaPtr;  // readout electronics
#ifdef USER_TUPLE
    static TNtupleD *fgTup; //! pointer to user ntuple
#endif

    ClassDef(Hit, 1)   // Hit class
};
}
#endif
