//=================================================================================
// Description:
// Script to plot spatial resolution as a function of drift length.
// You need to create input data ("sg.dat" by default) by running Resol.C.
//   isz = (true, false) = (z, r-phi)
//   isg = false switches off graphics, to be used when you want to overlay
//         the resolution curve on test beam data, etc.
//
// Example:
// To run, do, for instance,
// $ root -l
// [0] .L Plot.C
// [1] Plot(); // for z resolution
// [2] Plot(false); // for r-phi resolution
// [3] Plot(false,"sg.2.dat",false)->Draw("c same"); // to overlay other data
// [4] Plot(false,"sg.3.dat",false)->Draw("c same"); // and more
//
// Note:
//    This version does not assume that data are pre-orderd in ascending z.
//=================================================================================

double Fun(const double *xv, const double *pv)
{
  double sg2 = pv[0]*pv[0]
	     + pv[3]*pv[3]*xv[0]/(pv[1]/cos(pv[2]));
  return sqrt(sg2);
}

TMultiGraph *Plot(bool isz = true, const char *fname = "sg.dat", bool isg = true)
{
   double xmin = 0.;
   double xmax = 600.;
   double ymin = 0.;
   double ymax = isz ? 0.6 : 0.35;

   auto mgp = new TMultiGraph;

   //--
   // Define helper classes
   //--
   // class to hold a single run data
   class Line {
   public:
      Line(double fi, double lam, double z, double dz, double sgx, double dsgx, double sgz, double dsgz)
          : _phi(fi), _lambda(lam), _z(z), _dz(dz), _sgx(sgx), _dsgx(dsgx), _sgz(sgz), _dsgz(dsgz)
      {
      }

      void Print() const
      {
         cerr << setw(5) << setprecision(3) << setiosflags(ios::fixed) << _phi    << " "
              << setw(5) << setprecision(3) << setiosflags(ios::fixed) << _lambda << " "
              << setw(4) << setprecision(0) << setiosflags(ios::fixed) << _z      << " "
              << setw(4) << setprecision(0) << setiosflags(ios::fixed) << _dz     << " "
              << setw(8) << setprecision(6) << setiosflags(ios::fixed) << _sgx    << " "
              << setw(8) << setprecision(6) << setiosflags(ios::fixed) << _dsgx   << " "
              << setw(8) << setprecision(6) << setiosflags(ios::fixed) << _sgz    << " "
              << setw(8) << setprecision(6) << setiosflags(ios::fixed) << _dsgz
              << endl;
      }

   public:
      double _phi;
      double _lambda;
      double _z;
      double _dz;
      double _sgx;
      double _dsgx;
      double _sgz;
      double _dsgz;
   };

   // class to hold multiple runs
   class Data: public vector<Line> {
   public:
      enum class Ekey {
          z,      // drift length [mm]
          phi,    // phi          [rad]
          lambda  // lambda       [rad]
      };

      // data sorting function by key
      void sort(Ekey key = Ekey::z)
      {
         switch (key) {
             case Ekey::phi:
                std::sort(begin(), end(),
                     [](const Line &x, const Line &y) { return x._phi < y._phi; });
                break;
             case Ekey::lambda:
                std::sort(begin(), end(),
                     [](const Line &x, const Line &y) { return x._lambda < y._lambda; });
                break;
             default:
             case Ekey::z:
                std::sort(begin(), end(),
                     [](const Line &x, const Line &y) { return x._z < y._z; });
         }
      }

      // function to fill arrays 
      void set(double *fi, double *lam, 
               double *z, double *dz, double *sgx, double *dsgx, double *sgz, double *dsgz,
               int ilo = -1, int ihi = -1)
      {
         int npt = size();
         ilo = ilo < 0 ? 0 : ilo < npt ? ilo : npt-1;
         ihi = ihi < 0 ? npt : ihi < npt ? ihi : npt;
         for (int i=0; i<ihi-ilo; i++) {
            fi  [i] = operator[](ilo+i)._phi;
            lam [i] = operator[](ilo+i)._lambda;
            z   [i] = operator[](ilo+i)._z;
            dz  [i] = operator[](ilo+i)._dz;
            sgx [i] = operator[](ilo+i)._sgx;
            dsgx[i] = operator[](ilo+i)._dsgx;
            sgz [i] = operator[](ilo+i)._sgz;
            dsgz[i] = operator[](ilo+i)._dsgz;
         }
      }

      void Print() const
      {
         for (Line line : *this) {
            line.Print();
	 }
      }
   };

   //--
   // Create a Data folder.
   //--
   Data idata;

   //--
   // Read in input data and put them into the folder.
   //--
   ifstream in(fname);
   double d1, d2, e1, e2, e3, e4, e5;
   while ((in >> d1 >> d2 >> e1 >> e2 >> e3 >> e4 >> e5)) {
      idata.push_back(Line(d1,d2,e1,0.,e2,e3,e4,e5));
   }
   int npt = idata.size();
   cerr << "Read " << npt << " points" << endl;

   //--
   // Sort data first in ascending order of phi.
   //--
   idata.sort(Data::Ekey::phi);
#if 0
   idata.Print();
#endif

   //--
   // Split data into different phi sets and create a graph for each of them
   //--
   cerr << "phi = ";
   int loc  = 0;
   int nphi = 0;
   double phis = -9999.;
   vector<Data> jdata;
   jdata.push_back(Data()); // prepare a container for the 1st data set
   for (Line line : idata) {
      auto phi = line._phi;
      if (phi == phis || !loc) {
         jdata[nphi].push_back(line); // still in the same data set, add this line
      }
      if (phi != phis || loc == npt-1) { // new phi or end of data
         phis = phi;
         if (loc != npt-1) cerr << phi << "  ";
         auto mpt = jdata[nphi].size();
         if (mpt && loc) {
            jdata[nphi].sort(Data::Ekey::z); // sort data in ascending order of drift length
            //--
            // Create a graph object, and add it to a multi-graph
            //--
            double fi[mpt], lm[mpt], z[mpt], dz[mpt], sgx[mpt], dsgx[mpt], sgz[mpt], dsgz[mpt];
            jdata[nphi].set(fi,lm,z,dz,sgx,dsgx,sgz,dsgz);
            TGraphErrors *grp = nullptr;
            if (!isz) grp = new TGraphErrors(mpt, z, sgx, dz, dsgx);
            else      grp = new TGraphErrors(mpt, z, sgz, dz, dsgz);
	    mgp->Add(grp);
	    if (loc < npt-1) jdata.push_back(Data()); // prepare a container for the next data set
	    nphi++;
            if (loc != npt-1) jdata[nphi].push_back(line); // don't forget to add this to the new set 
	 }
      }
      loc++;
   }
   cerr << " --> found " << nphi << " phi points " << endl; 

   //--
   // Prepare a canvas if requested.
   //--
   TCanvas *c1 = nullptr;
   if (isg) {
      c1 = new TCanvas("c1", "c1",9,31,699,550);
      gStyle->SetOptStat(0);
      c1->ToggleEventStatus();
      c1->SetGridx();
      c1->SetGridy();
      c1->SetBorderSize(2);
      c1->SetLeftMargin(0.1280576);
      c1->SetBottomMargin(0.1343612);

      auto framep = mgp->GetHistogram();
      auto &frame = *framep;
      frame.SetStats(0);
      frame.GetXaxis()->SetLimits(xmin,xmax);
      frame.SetMinimum(ymin);
      frame.SetMaximum(ymax);
      frame.GetXaxis()->SetTitle("drift length [mm]");
      frame.GetXaxis()->SetLabelSize(0.05);
      frame.GetXaxis()->SetTitleSize(0.06);
      if (!isz) frame.GetYaxis()->SetTitle("#sigma_{x} [mm]");
      else      frame.GetYaxis()->SetTitle("#sigma_{z} [mm]");
      frame.GetYaxis()->SetLabelSize(0.05);
      frame.GetYaxis()->SetTitleSize(0.06);
      frame.GetYaxis()->SetTitleOffset(1.0);

      mgp->Draw("apc");

  /////FIXME////////////////////////////
      const double kCd       = (isz ? 0.220 : 0.096)/sqrt(10.); // [mm/sqrt[cm]
      const double xfitmin   = 150.;
      const double kRadToDeg = 180./TMath::Pi();
  /////FIXME////////////////////////////

      double xtaglo = xmin + 0.7*(xmax-xmin);

      auto glistp = mgp->GetListOfGraphs();
      TIter next(glistp);
      int iphi = 0;
      while (auto gp = static_cast<TGraph *>(next())) {
         auto ytag   = gp->Eval(xtaglo);
	 auto ytaglo = ytag - (ymax-ymin)*0.04;

         auto tagp = new TLatex;
         tagp->SetTextFont(132);
         tagp->SetTextAlign(12);
	 auto phi = jdata[iphi].operator[](0)._phi;
         stringstream stext;
#if 0
         stext << "#phi = " << setw(6) << setprecision(3) << phi << ends;
#else
         stext << "#phi = " << setw(6) << setprecision(3) << phi*kRadToDeg << "#circ" << ends;
#endif
         tagp->DrawLatex(xtaglo,ytaglo,stext.str().data());
	 tagp->Draw();

         auto funp = new TF1("fun",&Fun,xmin,xmax,4);
         funp->SetParameter(0,0.1);  // [A] term [mm]
         funp->SetParameter(1,30.);  // Neff
         funp->FixParameter(2,phi);  // phi [rad]
         funp->FixParameter(3,kCd);  // Cd [mm/sqrt[mm])
         //gp->Fit(funp,"Q","",-5,5);
         gp->Fit(funp,"","",xfitmin, xmax);
#if 1
         cerr << "------------------------------------------ " << endl;
         cerr << setprecision(1) << setiosflags(ios::fixed)
              << "phi = " <<  phi*kRadToDeg << " [deg]"
              << setprecision(2)
              << " Neff(F) = " << funp->GetParameter(1) 
              << " +/- "       << funp->GetParError (1)
              << endl;
         cerr << "------------------------------------------ " << endl;
#endif
	 iphi++;
      }
   }

   //--
   // Return the pointer to this multi-graph.
   //--
   return mgp;
}
